"""Comprehensive tests for the runner module."""

import pytest
from pathlib import Path
from lambdora.runner import run_file, load_std
from lambdora.errors import TokenizeError
from unittest.mock import patch

def test_run_file_error_handling():
    """Test run_file error handling."""
    # Test with non-existent file
    with pytest.raises(SystemExit):
        run_file(Path("nonexistent_file.lamb"))

def test_runner_import():
    """Test that runner module can be imported and functions exist."""
    assert callable(run_file)
    assert callable(load_std)

def test_load_std_basic():
    """Test basic load_std functionality."""
    # Should not raise an error
    load_std()

# Additional tests for missing coverage

def test_runner_error_conditions():
    """Test runner error conditions."""
    # Test runner with empty file
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("")
        f.flush()
        try:
            result = run_file(Path(f.name))
            assert result is None
        finally:
            os.unlink(f.name)

def test_runner_with_valid_file():
    """Test runner with valid file."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("(+ 1 2)")
        f.flush()
        try:
            result = run_file(Path(f.name))
            assert result is None  # run_file returns None but prints the result
        finally:
            os.unlink(f.name)

def test_runner_with_multiple_expressions():
    """Test runner with multiple expressions in file."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("(define x 42)\n(+ x 1)")
        f.flush()
        try:
            result = run_file(Path(f.name))
            assert result is None  # run_file returns None but prints the result
        finally:
            os.unlink(f.name)

def test_runner_with_comments():
    """Test runner with comments in file."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("; This is a comment\n(+ 1 2)")
        f.flush()
        try:
            result = run_file(Path(f.name))
            assert result is None  # run_file returns None but prints the result
        finally:
            os.unlink(f.name)

def test_runner_with_whitespace():
    """Test runner with whitespace in file."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("  \n  (+ 1 2)  \n  ")
        f.flush()
        try:
            result = run_file(Path(f.name))
            assert result is None  # run_file returns None but prints the result
        finally:
            os.unlink(f.name)

def test_runner_with_unicode_decode_error():
    """Test runner with UnicodeDecodeError."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='wb', suffix='.lamb', delete=False) as f:
        f.write(b'\xff\xfe\x00')  # Invalid UTF-8
        f.flush()
        try:
            with pytest.raises(SystemExit):
                run_file(Path(f.name))
        finally:
            os.unlink(f.name)


def test_runner_with_os_error():
    """Test runner with OSError."""
    with pytest.raises(SystemExit):
        run_file(Path("/nonexistent/path/file.lamb"))


def test_load_std_with_custom_path_not_found():
    """Test load_std with custom stdlib path that doesn't exist."""
    import tempfile
    import os
    with tempfile.TemporaryDirectory() as temp_dir:
        custom_std = Path(temp_dir) / "nonexistent.lamb"
        # Should not raise an error, should fall back to built-in
        load_std(custom_std)


def test_load_std_with_custom_path_exists():
    """Test load_std with custom stdlib path that exists."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("(define test-var 42)")
        f.flush()
        try:
            # Should not raise an error
            load_std(Path(f.name))
        finally:
            os.unlink(f.name)


def test_load_std_with_builtin_not_found():
    """Test load_std when built-in stdlib doesn't exist."""
    # This is a bit tricky to test, but we can mock the path
    with patch('pathlib.Path.exists', return_value=False):
        # Should not raise an error
        load_std()


def test_run_file_with_lamb_error():
    """Test run_file with LambError."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("(undefined-function 1 2 3)")
        f.flush()
        try:
            with pytest.raises(SystemExit):
                run_file(Path(f.name))
        finally:
            os.unlink(f.name)


def test_run_file_with_unexpected_exception():
    """Test run_file with unexpected exception."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("(+ 1 2)")
        f.flush()
        try:
            # Mock lambTokenize to raise an unexpected exception
            with patch('lambdora.runner.lambTokenize', side_effect=RuntimeError("Unexpected")):
                with pytest.raises(SystemExit):
                    run_file(Path(f.name))
        finally:
            os.unlink(f.name)


def test_run_file_with_macro_expansion_none():
    """Test run_file when macro expansion returns None."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("(define x 42)")
        f.flush()
        try:
            # Mock lambMacroExpand to return None
            with patch('lambdora.runner.lambMacroExpand', return_value=None):
                result = run_file(Path(f.name))
                assert result is None
        finally:
            os.unlink(f.name)


def test_run_file_with_output():
    """Test run_file with output that should be printed."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("(+ 1 2)")
        f.flush()
        try:
            with patch('builtins.print') as mock_print:
                result = run_file(Path(f.name))
                assert result is None
                mock_print.assert_called_once()
        finally:
            os.unlink(f.name)


def test_run_file_with_nil_output():
    """Test run_file with nil output (should not print)."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("(define x 42)")
        f.flush()
        try:
            with patch('builtins.print') as mock_print:
                result = run_file(Path(f.name))
                assert result is None
                mock_print.assert_not_called()
        finally:
            os.unlink(f.name)


def test_runner_main_function():
    """Test the legacy main function."""
    with patch('sys.argv', ['runner.py', 'test.lamb']):
        with patch('lambdora.runner.run_file') as mock_run_file:
            from lambdora.runner import main
            main()
            mock_run_file.assert_called_once()


def test_runner_main_function_wrong_args():
    """Test the legacy main function with wrong number of arguments."""
    with patch('sys.argv', ['runner.py']):
        with patch('sys.stderr') as mock_stderr:
            with pytest.raises(SystemExit):
                from lambdora.runner import main
                main()
            mock_stderr.write.assert_called()


def test_runner_main_function_too_many_args():
    """Test the legacy main function with too many arguments."""
    with patch('sys.argv', ['runner.py', 'file1.lamb', 'file2.lamb']):
        with patch('sys.stderr') as mock_stderr:
            with pytest.raises(SystemExit):
                from lambdora.runner import main
                main()
            mock_stderr.write.assert_called()


def test_load_std_with_lamb_error():
    """Test load_std with LambError during loading."""
    # Mock lambTokenize to raise a LambError
    with patch('lambdora.runner.lambTokenize', side_effect=TokenizeError("Test error")):
        with pytest.raises(SystemExit):
            load_std()


def test_run_file_with_multiple_expressions_output():
    """Test run_file with multiple expressions that produce output."""
    import tempfile
    import os
    with tempfile.NamedTemporaryFile(mode='w', suffix='.lamb', delete=False) as f:
        f.write("(+ 1 2)\n(+ 3 4)")
        f.flush()
        try:
            with patch('builtins.print') as mock_print:
                result = run_file(Path(f.name))
                assert result is None
                assert mock_print.call_count == 2
        finally:
            os.unlink(f.name)



"""Main CLI entry point for Lambdora."""

import argparse
import sys
from pathlib import Path
from typing import Optional

from . import __version__
from .repl import repl
from .runner import run_file


def create_parser() -> argparse.ArgumentParser:
    """Create the main argument parser."""
    parser = argparse.ArgumentParser(
        prog="lambdora",
        description="A minimalist Lisp-inspired functional language",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  lambdora repl                    # Start interactive REPL
  lambdora run script.lamb         # Execute a Lambdora script
  lambdora --version               # Show version information
  lambdora repl --stdlib-path /path/to/std.lamb  # Use custom stdlib
        """,
    )

    parser.add_argument(
        "--version", action="version", version=f"Lambdora {__version__}"
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # REPL subcommand
    repl_parser = subparsers.add_parser("repl", help="Start the interactive REPL")
    repl_parser.add_argument(
        "--stdlib-path",
        type=Path,
        help="Path to custom standard library file (default: built-in std.lamb)",
    )

    # Run subcommand
    run_parser = subparsers.add_parser("run", help="Execute a Lambdora script")
    run_parser.add_argument("file", help="Path to the .lamb file to execute")
    run_parser.add_argument(
        "--stdlib-path",
        type=Path,
        help="Path to custom standard library file (default: built-in std.lamb)",
    )

    return parser


def main(args: Optional[list[str]] = None) -> int:
    """Main CLI entry point."""
    parser = create_parser()
    parsed_args = parser.parse_args(args)

    try:
        if parsed_args.command == "repl":
            repl(stdlib_path=parsed_args.stdlib_path)
            return 0
        elif parsed_args.command == "run":
            file_path = Path(parsed_args.file)
            if not file_path.exists():
                print(f"Error: File '{file_path}' not found.", file=sys.stderr)
                print(
                    "Tip: Make sure the file exists and the path is correct.",
                    file=sys.stderr,
                )
                return 1
            if not file_path.suffix == ".lamb":
                print(
                    f"Warning: File '{file_path}' doesn't have .lamb extension.",
                    file=sys.stderr,
                )
                print(
                    f"Tip: Consider renaming to '{file_path.with_suffix('.lamb')}'",
                    file=sys.stderr,
                )
            run_file(file_path, stdlib_path=parsed_args.stdlib_path)
            return 0
        else:
            # No subcommand provided, show help
            parser.print_help()
            return 0
    except KeyboardInterrupt:
        print("\nInterrupted by user.", file=sys.stderr)
        return 1
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


if __name__ == "__main__":
    sys.exit(main())



from lambdora.astmodule import Application, Literal, Variable
from lambdora.repl import run_expr as runExpression


def test_quote_literal():
    result = runExpression("'42")
    assert isinstance(result, Literal)
    assert result.value == "42"  # returned as string, unevaluated


def test_quote_expr():
    result = runExpression("'(+ 1 2)")
    assert isinstance(result, Application)
    assert isinstance(result.func, Variable)
    assert result.func.name == "+"


def test_quote_prevents_eval():
    result = runExpression("'(+ 1 2)")
    assert result != 3


def test_quasiquote_unquote():
    runExpression("(define x 99)")
    result = runExpression("(quasiquote (+ 1 (unquote x)))")
    assert isinstance(result, Application)
    assert result.func.name == "+"
    assert result.args[1] == 99



"""Tokenizer for Lambdora source code (returns ``list[str]``)."""

from __future__ import annotations

from .errors import TokenizeError


def _line_at(src: str, line_no: int) -> str:
    """Return the given 1-based line from *src*."""

    return src.splitlines()[line_no - 1]


def lambTokenize(source: str, *, filename: str | None = None) -> list[str]:
    """Tokenise *source*. *filename* is used only in error messages."""

    tokens: list[str] = []
    i = 0  # absolute index into *source*
    line_no = 1
    col_no = 1  # 1-based column index

    while i < len(source):
        char = source[i]

        # Newline
        if char == "\n":
            i += 1
            line_no += 1
            col_no = 1
            continue

        # Skip ';' comments
        if char == ";":
            while i < len(source) and source[i] != "\n":
                i += 1
                col_no += 1
            continue  # newline (if any) handled on next loop iteration

        # Whitespace
        if char.isspace():
            i += 1
            col_no += 1
            continue

        # Multi-char operators (check before single-char tokens)
        if i + 1 < len(source):
            two_char = source[i : i + 2]
            if two_char in ["++", "!=", "<=", ">="]:
                tokens.append(two_char)
                i += 2
                col_no += 2
                continue

        # Single-char tokens
        if char in "().+-*/%=<>',`":
            tokens.append(char)
            i += 1
            col_no += 1
            continue

        # Identifiers
        if char.isalpha() or char == "_":
            start = i
            while i < len(source) and (
                source[i].isalnum() or source[i] == "_" or source[i] == "-" or source[i] == "?"
            ):
                i += 1
                col_no += 1
            tokens.append(source[start:i])
            continue

        # Integers
        if char.isdigit():
            start = i
            while i < len(source) and source[i].isdigit():
                i += 1
                col_no += 1
            tokens.append(source[start:i])
            continue

        # Strings
        if char == '"':
            i += 1
            col_no += 1
            start_idx = i
            str_line, str_col = line_no, col_no

            while i < len(source) and source[i] != '"':
                if source[i] == "\n":
                    line_no += 1
                    col_no = 0  # will be incremented at end of loop
                i += 1
                col_no += 1

            if i >= len(source):  # reached EOF
                snippet = _line_at(source, str_line)
                raise TokenizeError(
                    "Unterminated string literal",
                    file=filename,
                    line=str_line,
                    column=str_col,
                    snippet=snippet,
                )

            # Slice out the contents (excluding the quotes)
            literal = source[start_idx:i]
            tokens.append(f'"{literal}"')

            # Skip the closing quote
            i += 1
            col_no += 1
            continue

        # Unknown char
        snippet = _line_at(source, line_no)
        raise TokenizeError(
            f"Unexpected character: {char}",
            file=filename,
            line=line_no,
            column=col_no,
            snippet=snippet,
        )

    return tokens



"""Runtime value representations used by the interpreter."""

from dataclasses import dataclass
from typing import Callable, List, Union

from .astmodule import Expr

Value = Union[
    int, str, bool, "Closure", "Builtin", "Pair", "Nil", "Macro", "Thunk", Expr
]


@dataclass
class Closure:
    param: str
    body: Expr
    env: dict[str, Value]


@dataclass
class Builtin:
    func: Callable[[Value], Value]


@dataclass
class Pair:
    head: Value
    tail: Value


@dataclass
class Macro:
    params: List[str]
    body: Expr


@dataclass
class Thunk:
    func: Callable[[], Value]


class Nil:
    def __repr__(self) -> str:
        return "nil"


nil = Nil()


def valueToString(val: Value) -> str:
    if isinstance(val, Closure):
        return f"<closure lambda {val.param}. …>"
    elif isinstance(val, bool):
        return "true" if val else "false"
    elif isinstance(val, int):
        return str(val)
    elif isinstance(val, str):
        return f"{val}"
    elif isinstance(val, Builtin):
        return "<builtin fn>"
    elif isinstance(val, Pair):
        # Print as (a b c)
        elems = []
        p: Value = val
        while isinstance(p, Pair):
            elems.append(valueToString(p.head))
            p = p.tail
        if p is not nil:
            elems.append(".")
            elems.append(valueToString(p))
        return f"({' '.join(elems)})"
    elif val is nil:
        return "nil"
    elif isinstance(val, Expr):
        # Handle AST nodes as values (code as data)
        from .printer import lambPrint

        return lambPrint(val)
    else:
        return f"<unknown value: {val}>"



"""AST node definitions for the Lambdora language."""

from dataclasses import dataclass
from typing import List


@dataclass
class Expr:
    pass


@dataclass
class Variable(Expr):
    name: str


@dataclass
class Literal(Expr):
    value: str


@dataclass
class Abstraction(Expr):
    param: str
    body: Expr


@dataclass
class Application(Expr):
    func: Expr
    args: List[Expr]


@dataclass
class DefineExpr(Expr):
    name: str
    value: Expr


@dataclass
class IfExpr(Expr):
    cond: Expr
    then_branch: Expr
    else_branch: Expr


@dataclass
class DefMacroExpr(Expr):
    name: str
    params: List[str]
    body: Expr


@dataclass
class QuoteExpr(Expr):
    value: Expr


@dataclass
class QuasiQuoteExpr(Expr):
    expr: Expr


@dataclass
class UnquoteExpr(Expr):
    expr: Expr


@dataclass
class LetRec(Expr):
    bindings: List[tuple[str, Expr]]
    body: List[Expr]



import pytest
from lambdora.repl import run_expr as runExpression
from lambdora.errors import RecursionInitError

def test_factorial_letrec():
    src = """
    (letrec ((fact (lambda n. (if (= n 0) 1 (* n (fact (- n 1)))))))
      (fact 5))
    """
    assert runExpression(src.strip()) == 120

def test_mutual_even_odd():
    src = """
    (letrec (
        (even (lambda n. (if (= n 0) true (odd (- n 1)))))
        (odd  (lambda n. (if (= n 0) false (even (- n 1)))))
      )
      (and (even 10) (not (even 11))))
    """
    assert runExpression(src.strip()) is True

def test_single_binding_recursion():
    expr = "(letrec ((fact (lambda n. (if (= n 0) 1 (* n (fact (- n 1))))))) (fact 6))"
    assert runExpression(expr) == 720

def test_placeholder_access_error():
    faulty_src = """
    (letrec (
       (x x))
       x)
    """
    with pytest.raises(RecursionInitError):
        runExpression(faulty_src.strip()) 